Attribute VB_Name = "modMain"
' //
' // Vectors normalization test using different approaches
' // by The trick
' //

Option Explicit

Private Const ERR_TOLERANCE     As Single = 0.01        ' // Error checking tolerance

Public Type tVector4D
    fX As Single
    fY As Single
    fZ As Single
    fW As Single
End Type

Private Declare Function QueryPerformanceCounter Lib "kernel32" ( _
                         ByRef lpPerformanceCount As Currency) As Long
Private Declare Function QueryPerformanceFrequency Lib "kernel32" ( _
                         ByRef lpFrequency As Currency) As Long

Private msLog   As String   ' // Events log

Sub Main()
    Dim tVec()      As tVector4D
    Dim tVecTemp()  As tVector4D
    Dim tDirect()   As tVector4D
    
    ReDim tVec(10000000)
    
    ' // Fill the vectors with different random values with range 0...10000
    FillRandom tVec
    
    tVecTemp = tVec
    
    LogEvent "NormalizeVectorsDirectVB", True
    
    ' // Default VB-implementation using direct sqr calculation
    NormalizeVectorsDirectVB tVecTemp
    
    LogEvent ""
    
    tDirect = tVecTemp
    
    tVecTemp = tVec
    
    LogEvent "NormalizeVectorsFastVB", True
    
    ' // Using VB fast inverse square root
    NormalizeVectorsFastVB tVecTemp

    LogEvent ""
    
    CheckError tDirect, tVecTemp
    
    tVecTemp = tVec
    
    LogEvent "NormalizeVectorsFastAsmFPU", True
    
    ' // Using ASM (FPU) fast inverse square root
    NormalizeVectorsFastAsmFPU tVecTemp

    LogEvent ""
    
    CheckError tDirect, tVecTemp
    
    tVecTemp = tVec
    
    LogEvent "NormalizeVectorsFastAsmSSE", True
    
    ' // Using ASM (SSE3) fast inverse square root
    NormalizeVectorsFastAsmSSE tVecTemp

    LogEvent ""
    
    CheckError tDirect, tVecTemp
    
    MsgBox msLog
    
End Sub

' //
' // Check error
' //
Private Sub CheckError( _
            ByRef tDirect() As tVector4D, _
            ByRef tVectors() As tVector4D)
    Dim lIndex  As Long
    
    For lIndex = 0 To UBound(tDirect)
        
        ' // 0.01 - tolerance
        If Abs(tDirect(lIndex).fX - tVectors(lIndex).fX) > ERR_TOLERANCE Or _
           Abs(tDirect(lIndex).fY - tVectors(lIndex).fY) > ERR_TOLERANCE Or _
           Abs(tDirect(lIndex).fZ - tVectors(lIndex).fZ) > ERR_TOLERANCE Or _
           Abs(tDirect(lIndex).fW - tVectors(lIndex).fW) > ERR_TOLERANCE Then
            Stop
        End If
        
    Next
    
End Sub

' //
' // Normalize array of vectors using direct calculatuion
' //
Private Sub NormalizeVectorsDirectVB( _
            ByRef tVectors() As tVector4D)
    Dim lIndex  As Long
    Dim fLen    As Single
    
    For lIndex = 0 To UBound(tVectors)
    
        fLen = Sqr(tVectors(lIndex).fX * tVectors(lIndex).fX + _
                   tVectors(lIndex).fY * tVectors(lIndex).fY + _
                   tVectors(lIndex).fZ * tVectors(lIndex).fZ + _
                   tVectors(lIndex).fW * tVectors(lIndex).fW)
                   
        tVectors(lIndex).fX = tVectors(lIndex).fX / fLen
        tVectors(lIndex).fY = tVectors(lIndex).fY / fLen
        tVectors(lIndex).fZ = tVectors(lIndex).fZ / fLen
        tVectors(lIndex).fW = tVectors(lIndex).fW / fLen
        
    Next
    
End Sub

' //
' // Normalize array of vectors using fast inverse square
' //
Private Sub NormalizeVectorsFastVB( _
            ByRef tVectors() As tVector4D)
    Dim lIndex  As Long
    Dim fLen    As Single
    Dim fHalf   As Single
    Dim lInt    As Long
    
    For lIndex = 0 To UBound(tVectors)
    
        fLen = tVectors(lIndex).fX * tVectors(lIndex).fX + _
               tVectors(lIndex).fY * tVectors(lIndex).fY + _
               tVectors(lIndex).fZ * tVectors(lIndex).fZ + _
               tVectors(lIndex).fW * tVectors(lIndex).fW
                   
        fHalf = fLen * 0.5!
        
        GetMem4 fLen, lInt
        
        lInt = &H5F3759DF - (lInt \ 2)
        
        GetMem4 lInt, fLen
        
        fLen = fLen * (1.5! - (fHalf * fLen * fLen))
        
        tVectors(lIndex).fX = tVectors(lIndex).fX * fLen
        tVectors(lIndex).fY = tVectors(lIndex).fY * fLen
        tVectors(lIndex).fZ = tVectors(lIndex).fZ * fLen
        tVectors(lIndex).fW = tVectors(lIndex).fW * fLen
        
    Next
    
End Sub

' // Log
Public Sub LogEvent( _
           Optional ByRef sText As String, _
           Optional ByVal bNoTime As Boolean)
    Static cFreq    As Currency, _
           cOldVal  As Currency
    Dim cNewVal     As Currency
    
    If cFreq = 0@ Then
    
        QueryPerformanceFrequency cFreq
        QueryPerformanceCounter cNewVal
        cOldVal = cNewVal
        
    Else
    
        QueryPerformanceCounter cNewVal
        
    End If
    
    If Not bNoTime Then
        msLog = msLog & Format$((cNewVal - cOldVal) / cFreq, "0.00000") & "ms. "
    End If
    
    msLog = msLog & sText & vbNewLine
        
    cOldVal = cNewVal
        
End Sub


